# -*- coding: UTF-8 -*-
"""
Nektar for Python
~~~~~~~~~~~~~~~~

Lightweight and highly customizable communication with the Hive APIs

"""

__version__ = "0.9.7"
from .nektar import Waggle
from .appbase import AppBase
__all__ = ["nektar", "appbase"]